"""
Django settings for picostack project.

For more information on this file, see
https://docs.djangoproject.com/en/1.6/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.6/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os
BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.6/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '-9@)*p*&1hk+6$2ura%9m8pf(sdz@us4p4!y%zq(#0meck2t+o'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = ['*']


# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'picostack.vms',
    'bootstrap3',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'picostack.urls'

WSGI_APPLICATION = 'picostack.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.6/ref/settings/#databases

# Use as location for the database file.
DATABASE_LOCATION = '/var/picostack/db/picostk.sqlite3'
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': DATABASE_LOCATION,
    }
}

# Internationalization
# https://docs.djangoproject.com/en/1.6/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.6/howto/static-files/

STATIC_URL = '/static/'

SETTINGS_PATH = os.path.normpath(os.path.dirname(__file__))
# Find templates in the same folder as settings.py.
TEMPLATE_DIRS = (
    os.path.join(SETTINGS_PATH, 'templates'),
)
STATICFILES_DIRS = (
    os.path.join(SETTINGS_PATH, 'static'),
)


BOOTSTRAP3 = {
    'jquery_url': STATIC_URL + 'bower_components/jquery/dist/jquery.min.js',
    'base_url': STATIC_URL + 'bower_components/bootstrap/dist/',
    'css_url': None,
    'theme_url': STATIC_URL + 'bower_components/bootstrap/dist/css/bootstrap-theme.min.css',
    'javascript_url': None,
    'horizontal_label_class': 'col-md-2',
    'horizontal_field_class': 'col-md-4',
}



import os
import sys
import unittest

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "picostack.settings")
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

from django.test import TestCase
from picostack.vms.models import Flavour, VmImage, VmInstance


class InstanceTestCase(TestCase):

    def setUp(self):
        vm_image = VmImage.objects.create(
            name='test_image',
            image_filename='test.img',
            disk_size=10,
        )
        vm_image.save()
        flavour = Flavour.objects.create(name='tiny')
        flavour.save()
        machine = VmInstance(
            name='test_vm',
            image=vm_image,
            flavour=flavour,
        )
        machine.save()

    def test_vm_basic(self):
        for machine in VmInstance.objects.all():
            #print machine.name
            assert machine.num_of_cores > 0

    def test_get_occupied_ports(self):
        # Make sure we have at least 1 mapped port.
        machine = VmInstance.objects.get(name='test_vm')
        machine.current_state = 'Running'
        machine.has_ssh = True
        machine.map_port('ssh', 10020)
        machine.save()
        # Test the routine.
        occupied_ports = VmInstance.get_all_occupied_ports()
        assert len(occupied_ports) > 0
        #print occupied_ports


if __name__ == "__main__":
    unittest.main()



import textwrap


def wrap_multiline(text, separator='\n'):
    compact_lines = [line.strip() for line in text.split('\n')]
    compacted_text = '\n'.join(compact_lines)
    command_lines = textwrap.wrap(compacted_text, width=210,
                                  break_on_hyphens=False,
                                  break_long_words=False)
    return separator.join(command_lines).strip()



import os
import signal
import shutil
import logging
import psutil
from collections import deque
from picostack.textwrap_util import wrap_multiline
from picostack.vms.models import (
    VmInstance, VM_PORTS,
    VM_IN_CLONING, VM_IS_STOPPED, VM_IS_LAUNCHED, VM_IS_RUNNING,
    VM_HAS_FAILED, VM_IS_TERMINATING, VM_IS_TRASHED,
)
from process_spawn import ProcessUtil

logger = logging.getLogger(__name__)


class VmManager(object):

    def __init__(self, config):
        self.config = config
        self.__next_unmapped_port = None

    @property
    def vm_image_path(self):
        return self.config.get('vm_manager', 'vm_image_path')

    @property
    def vm_disk_path(self):
        return self.config.get('vm_manager', 'vm_disk_path')

    def validate_config(self):
        assert self.config.has_section('vm_manager')
        assert os.path.exists(self.vm_image_path)
        assert os.path.exists(self.vm_disk_path)

    @property
    def mapping_port_range(self):
        first_port = int(self.config.get('app', 'first_mapped_port'))
        last_port = int(self.config.get('app', 'last_mapped_port'))
        assert last_port > first_port
        if self.__next_unmapped_port is None \
                or self.__next_unmapped_port > last_port:
            self.__next_unmapped_port = first_port
        mapping_port_range = range(first_port, last_port)
        result = deque(mapping_port_range)
        shift = self.__next_unmapped_port - first_port + 1
        result.rotate(-1 * shift)
        return result

    def get_next_unmapped_port(self):
        '''
        Get a next port form the mapping range that was not mapped by any
        instances.
        '''
        # Get a list of ports, occupied by running instances
        already_mapped_ports = VmInstance.get_all_occupied_ports()
        # Continue until unmapped port is found.
        for next_port in self.mapping_port_range:
            if next_port in already_mapped_ports:
                continue
            # Found unmapped port.
            self.__next_unmapped_port = next_port
            return next_port
        raise Exception('Failed to find unmapped/unoccupied port.')

    @property
    def location_of_images(self):
        return self.config.get('vm_manager', 'vm_image_path')

    def get_image_path(self, image):
        return os.path.join(self.location_of_images, image.image_filename)

    @property
    def location_of_disks(self):
        return self.config.get('vm_manager', 'vm_disk_path')

    def get_disk_path(self, machine):
        return os.path.join(self.location_of_disks, machine.disk_filename)

    def get_pid_file(self, machine):
        pidfiles_folder = self.config.get('app', 'pidfiles_path')
        return os.path.join(pidfiles_folder, '%s.pid' % machine.name)

    def get_report_file(self, machine):
        logfiles_folder = self.config.get('app', 'log_path')
        return os.path.join(logfiles_folder, '%s.log' % machine.name)

    @classmethod
    def create(self, name, config):
        '''Fabric of VM managers'''
        if name.upper() == 'KVM':
            return Kvm(config)
        raise Exception('Unknown VM manager: %s' % name)

    def build_machines(self):
        instances = VmInstance.objects.filter(current_state=VM_IN_CLONING)
        if not instances.exists():
            logger.info('Nothing to clone..')
            return
        for machine in instances:
            logger.info('Cloning "%s"' % machine.name)
            self.clone_from_image(machine)

    def start_machines(self):
        instances = VmInstance.objects.filter(current_state=VM_IS_LAUNCHED)
        if not instances.exists():
            logger.info('Nothing to start..')
            return
        for machine in instances:
            logger.info('Start running machine "%s"' % machine.name)
            self.run_machine(machine)

    def stop_machines(self):
        instances = VmInstance.objects.filter(current_state=VM_IS_TERMINATING)
        if not instances.exists():
            logger.info('Nothing to stop..')
            return
        for machine in instances:
            logger.info('Terminating machine "%s"' % machine.name)
            self.stop_machine(machine)

    def destory_machines(self):
        instances = VmInstance.objects.filter(current_state=VM_IS_TRASHED)
        if not instances.exists():
            logger.info('Nothing to trash..')
            return
        for machine in instances:
            logger.info('Trashing machine "%s"' % machine.name)
            self.remove_machine(machine)

    def run_machine(self, machine):
        raise NotImplementedError()

    def stop_machine(self, machine):
        raise NotImplementedError()

    def clone_from_image(self, machine):
        raise NotImplementedError()

    def remove_machine(self, vm_image):
        raise NotImplementedError()

    def kill_all_machines(self):
        raise NotImplementedError()


class Kvm(VmManager):

    def get_kvm_call(self, machine):
        # Make a list of ports to redirect from the VM to host. Ports will be
        # available at the host computer.
        redirected_ports = ''
        ports_to_map = list()
        if machine.has_ssh:
            ports_to_map.append('ssh')
        if machine.has_vnc:
            # TODO: check if VNC should be a "redirected port"?
            ports_to_map.append('vnc')
        if machine.has_rdp:
            ports_to_map.append('rdp')
        for port_to_map in ports_to_map:
            unmapped_port = self.get_next_unmapped_port()
            machine.map_port(port_to_map, unmapped_port)
            redirected_ports += ' -redir tcp:%d::%d ' % (unmapped_port,
                                                         VM_PORTS[port_to_map])
        host_vnc = '-vnc localhost:%d' % machine.localhost_vnc_port
        # Make a command line text with KVM call.
        return wrap_multiline('''
            /usr/bin/kvm -machine accel=kvm -hda %(disk_path)s
                -boot c
                -m %(memory_size)s
                -cpu qemu64 -smp %(num_of_cores)s,cores=%(num_of_cores)s,sockets=1,threads=1
                -net nic,model=virtio -net user
                %(redirected_ports)s
                -usbdevice tablet
                %(host_vnc)s
        ''' % {
            'disk_path': self.get_disk_path(machine),
            'memory_size': machine.memory_size,
            'num_of_cores': machine.num_of_cores,
            'redirected_ports': redirected_ports,
            'host_vnc': host_vnc,
        }, separator=' ')

    def run_machine(self, machine):
        # Check if machine is in accepting state.
        assert machine.current_state == VM_IS_LAUNCHED
        # Bake a shell command to spawn the machine.
        shell_command = self.get_kvm_call(machine)
        logger.debug('Running VM with shell command:\n%s' % shell_command)
        #output = invoke(command)
        report_filepath = self.get_report_file(machine)
        pid_filepath = self.get_pid_file(machine)
        assert not ProcessUtil.process_runs(pid_filepath)
        ProcessUtil.exec_process(shell_command, report_filepath, pid_filepath)
        # Update state.
        machine.change_state(VM_IS_RUNNING)

    def stop_machine(self, machine):
        # Check if machine is in accepting state.
        assert machine.current_state == VM_IS_TERMINATING
        # Kill the machine by pid.
        cxt_pidfile_filepath = self.get_pid_file(machine)
        proc_pidfile_path = '%s_proc' % cxt_pidfile_filepath
        # First kill proc which is a child and then daemoncxt.
        if ProcessUtil.kill_process(proc_pidfile_path) \
                and ProcessUtil.kill_process(cxt_pidfile_filepath):
            logging.info('Succefully stoping VM processes as in %s and %s' %
                         (proc_pidfile_path, cxt_pidfile_filepath))
            # Proc pid should be taken care of.
            if os.path.exists(proc_pidfile_path):
                os.unlink(proc_pidfile_path)
        else:
            logging.warning('Expected VM process does not run anymore. '
                            'Please check the log file for details: %s' %
                            self.get_report_file(machine))
        # Update state.
        machine.change_state(VM_IS_STOPPED)

    def clone_from_image(self, machine):
        # Check if machine is in accepting state.
        assert machine.current_state == VM_IN_CLONING
        logger.info('Cloning new machine \'%s\' form image \'%s\'' %
                    (machine.name, machine.image.name))
        # Copy machine. Can take time.
        src_file = self.get_image_path(machine.image)
        dst_file = self.get_disk_path(machine)
        logger.info('Copying %s -> %s' %
                    (src_file, dst_file))
        shutil.copyfile(src_file, dst_file)
        # Update state to VM_IS_STOPPED - we are ready to run.
        machine.change_state(VM_IS_STOPPED)

    def remove_machine(self, machine):
        # Check if machine is in accepting state.
        assert machine.current_state == VM_IS_TRASHED
        logger.info('Removing trashed machine \'%s\' and its files: \'%s\'' %
                    (machine.name, machine.disk_filename))
        disk_file = self.get_disk_path(machine)
        os.unlink(disk_file)
        # Clean logs.
        report_filepath = self.get_report_file(machine)
        if os.path.exists(report_filepath):
            os.unlink(report_filepath)
        # Finally kill the DB record.
        machine.delete()

    def kill_all_machines(self):
        needle = self.vm_disk_path
        machines = list()
        for proc in psutil.process_iter():
            try:
                pinfo = proc.as_dict(attrs=['pid', 'cmdline'])
            except psutil.NoSuchProcess:
                pass
            else:
                if needle in ' '.join(pinfo['cmdline']):
                    machines.append(pinfo)
        for machine in machines:
            os.kill(machine['pid'], signal.SIGTERM)




'''
Run on a local linux machine a subprocess in a daemon mode. Log the result.
'''
import os
import sys
import time
import signal
import logging
from StringIO import StringIO
from datetime import datetime
from daemoncxt.daemon import DaemonContext
import errno
from daemoncxt.lockfile import LockTimeout
from daemoncxt.pidlockfile import TimeoutPIDLockFile
from subprocess import (PIPE, Popen)


logger = logging.getLogger(__name__)


class ExecProcessError(Exception):
    '''Raised if forking error or similar problem occurred.'''


NEW_LINE = '\n'
NUM_SECTION_LINES = 1000
HZ = os.sysconf(os.sysconf_names['SC_CLK_TCK'])
SPAWN_TRIES = 3


def invoke(command, _in=None):
    '''
    Invoke command as a new system process and return its output.
    '''
    process = Popen(command, stdin=PIPE, stdout=PIPE, shell=True,
                    executable='/bin/bash')
    if _in is not None:
        process.stdin.write(_in)
    return process.stdout.read()


def strfdelta(tdelta, fmt):
    d = {"days": tdelta.days}
    d["hours"], rem = divmod(tdelta.seconds, 3600)
    d["minutes"], d["seconds"] = divmod(rem, 60)
    return fmt.format(**d)

LOCAL_TIME_FMT = '{days} days {hours}:{minutes}:{seconds}'


class ProcessUtil(object):

    __boot_time = None

    @classmethod
    def get_boot_time(cls):
        if cls.__boot_time is None:
            system_stats = open('/proc/stat').readlines()
            for line in system_stats:
                if line.startswith('btime'):
                    cls.__boot_time = int(line.split()[1])
        return cls.__boot_time

    @classmethod
    def get_birthtime_secs(cls, pid):
        process_stats = open('/proc/%d/stat' % int(pid)).read().split()
        age_from_boot_jiffies = int(process_stats[21])
        age_from_boot_timestamp = age_from_boot_jiffies / HZ
        age_timestamp = cls.get_boot_time() + age_from_boot_timestamp
        return age_timestamp

    @classmethod
    def list_processes(cls):
        '''See also man 5 proc'''
        pids = [pid for pid in os.listdir('/proc') if pid.isdigit() \
                and os.access(os.path.join('/proc', pid, 'status'), os.R_OK) \
                and os.access(os.path.join('/proc', pid, 'cmdline'), os.R_OK)]
        for pid in pids:
            proc_info = dict(id=pid)
            try:
                proc_info['cmdline'] = open(os.path.join('/proc', pid, 'cmdline'), 'rb').read()
                #proc_info['cwd'] = os.readlink(os.path.join('/proc', pid, 'cwd'))
                proc_info['running_since'] = ProcessUtil.get_birthtime_secs(pid)
                # Read process status
                proc_status = open(os.path.join('/proc', pid, 'status'), 'rb').read()
                for line in proc_status.split('\n'):
                    if not ':' in line:
                        continue
                    key, value = line.split(':', 1)
                    proc_info[key.strip()] = value.strip()
                yield proc_info
            except IOError:
                logger.debug('Can not read process (%d). Skipping..', pid)

    @classmethod
    def exec_process(cls, shell_command, report_filename, pidfile_path):

        def report_contains_error(report_filename):
            # Implement parsing. If no error happend file will exists but will
            # be empty. If things happens very fast but still ok  than the
            # section after 'Process stderror was:\n' token will be empty.
            return False

        def fork_parent(shell_command, report_filename, pidfile_path,
                        error_message):
            """ Fork a child process.

                If the fork fails, raise a ``DaemonProcessDetachError``
                with ``error_message``.
                """
            try:
                pid = os.fork()
                if pid > 0:
                    logger.debug('Waiting for pidfile which means process '
                                 'was spawned.')
                    os.waitpid(pid, 0)
                    if os.path.exists(pidfile_path) \
                            and os.path.exists(report_filename):
                        logger.debug('Giving it another try while waiting '
                                     'for pidfile..')
                        # Wait and parse stderr output in the log file.
                        # Assume process spawn only if such output is empty.
                        if not report_contains_error(report_filename):
                            logger.debug(
                                'Process was successfully spawned as %d' %
                                pid
                            )
                            return pid
                    # Give it even more chances.. Waiting for the pidfile.
                    for try_num in xrange(SPAWN_TRIES):
                        logger.debug('..retrying')
                        if os.path.exists(pidfile_path) \
                                and os.path.exists(report_filename):
                            # Repeat parsing effort of the log file.
                            if not report_contains_error(report_filename):
                                logger.debug(
                                    'Process was successfully spawned as %d' %
                                    pid
                                )
                                return pid
                        time.sleep(1)
                    raise ExecProcessError('Failed to spawn process '
                                           'after (%d) retries..' %
                                           SPAWN_TRIES)
                else:
                    with DaemonContext(
                        pidfile=ProcessUtil.get_pidfile(pidfile_path),
                        stdout=sys.stdout,
                        stderr=sys.stderr,
                        detach_process=True,
                    ) as process:
                        process_error = ''
                        # Write report file.
                        report = open(report_filename, 'a+')
                        started_at = datetime.now()
                        success = True
                        try:
                            cmd_args = [arg for arg in shell_command.split(' ')
                                        if len(arg) > 0]
                            proc = Popen(cmd_args, stdout=PIPE, stderr=PIPE)
                            # Second pid of submissive process, that does the
                            # actual work. Save it so it can be killed as well
                            proc_pidfile_path = '%s_proc' % pidfile_path
                            with open(proc_pidfile_path, 'w+') as proc_pidfile:
                                proc_pidfile.write('%d' % proc.pid)
                            # Do actual call.
                            process_output, process_error = proc.communicate()
                            # TODO: check return code?
                        except Exception as exception:
                            stderr = StringIO()
                            process_error = repr(exception)
                            exc_type, exc_value, exc_traceback = sys.exc_info()
                            # traceback.print_exception(exc_type, exc_value,
                            #                           exc_traceback,
                            #                           limit=3, file=stderr)
                            process_error += '\n' + stderr.getvalue().strip()
                            print process_error
                            success = False
                        time.sleep(5)
                        elapsed = datetime.now() - started_at
                        # TODO: gather /proc stats for current process
                        report.write('Elapsed time: %s \n' %
                                     strfdelta(elapsed, LOCAL_TIME_FMT))
                        report.write('Your job looked like:\n')
                        report.write(shell_command + '\n')
                        if success:
                            report.write('Successfully completed.\n')
                        else:
                            report.write('Job failed with an error: %s.\n' %
                                         process_error)
                        report.write('The output (if any) follows:\n')
                        report.write('Process output was:\n')
                        report.write(process_output)
                        report.write('Process stderror was:\n')
                        report.write(process_error)
                        report.close()
                    # Exit child after work is done.
                    exit(0)
            except OSError, exc:
                exc_errno = exc.errno
                exc_strerror = exc.strerror
                error = ExecProcessError(
                    "%(error_message)s: [%(exc_errno)d] %(exc_strerror)s" %
                    vars())
                raise error
        # Main part of exec_process()
        try:
            if os.path.exists(pidfile_path):
                raise ExecProcessError('Error! A pidfile already exists: ' +
                                       pidfile_path)
            pid = fork_parent(shell_command, report_filename, pidfile_path,
                              error_message="Failed to fork")
            # TODO: think of using os.setsid()?
            return pid
        except LockTimeout:
            raise ExecProcessError('Process pidfile is locked: ' +
                                   pidfile_path)

    @staticmethod
    def get_pidfile(pidfile_path):
        '''Note that pidfile_path must be absolute'''
        return TimeoutPIDLockFile(
            pidfile_path,
            acquire_timeout=5,
            threaded=False)

    @classmethod
    def pid_exists(cls, pid):
        """Check whether pid exists in the current process table."""
        if pid < 0:
            return False
        try:
            os.kill(pid, 0)
        except OSError, e:
            return e.errno == errno.EPERM
        else:
            return True

    @classmethod
    def process_runs(cls, pidfile_path):
        logger.debug('Process runs? ' + pidfile_path)
        if os.path.exists(pidfile_path):
            try:
                pid = int(open(pidfile_path).read().strip())
                if cls.pid_exists(pid):
                    return True
            except ValueError:
                return False
        return False

    @classmethod
    def kill_process(cls, pidfile_path):
        if not os.path.exists(pidfile_path):
            return False
        pid = int(open(pidfile_path).read())
        if not ProcessUtil.process_runs(pidfile_path):
            logger.warn('Proccess was not found %d' % pid)
            return False
        # SIGTERM is handled by DaemonContext
        os.kill(pid, signal.SIGTERM)
        return True

    @staticmethod
    def kill_process_pid(pid):
        os.kill(pid, signal.SIGTERM)




import os
import logging
import textwrap
from picostack.socket_logger import LogRecordSocketReceiver


WHITE_LIST = ['picostk', 'picostack']
VERBOSITY_LEVELS = {
    1: logging.WARN,
    2: logging.INFO,
    3: logging.DEBUG,
    4: logging.NOTSET,
}
LOGGING_LEVELS = {
    'WARN': logging.WARN,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
    'NOTSET': logging.NOTSET,
}
VERBOSITY_TO_LEVELS = {
    1: 'WARN',
    2: 'INFO',
    3: 'DEBUG',
    4: 'NOTSET',
}


class Whitelist(logging.Filter):

    def __init__(self, *whitelist):
        self.whitelist = [logging.Filter(name) for name in whitelist]

    def filter(self, record):
        return any(f.filter(record) for f in self.whitelist)


def set_interactive_logging(verbosity_level):
    logging.basicConfig(
        level=VERBOSITY_LEVELS[verbosity_level],
        format='%(asctime)s %(name)-20s %(levelname)-8s %(message)s',
        datefmt='%m-%d %H:%M')
    for handler in logging.root.handlers:
        handler.addFilter(Whitelist(*WHITE_LIST))


def set_logging_as_socket_client():
    rootLogger = logging.getLogger('')
    rootLogger.setLevel(logging.DEBUG)
    socketHandler = logging.handlers.SocketHandler(
        'localhost', logging.handlers.DEFAULT_TCP_LOGGING_PORT)
    socketHandler.addFilter(Whitelist(*WHITE_LIST))
    rootLogger.addHandler(socketHandler)


def set_logging_as_socket_server(logging_config_filename):
    logging.config.fileConfig(logging_config_filename,
                              disable_existing_loggers=False)
    # logging.basicConfig(
    #     format='%(relativeCreated)5d %(name)-15s %(levelname)-8s %(message)s')
    tcpserver = LogRecordSocketReceiver()
    tcpserver.serve_until_stopped()


def fork_me_socket_logging(logging_config_filename):
    # Fork me a server
    try:
        pid = os.fork()
        if pid == 0:
            # I am a child and will become a server.
            set_logging_as_socket_server(logging_config_filename)
            # Exit child and another after work is done.
            exit(0)
        else:
            # I am a parent and will emit as a client.
            set_logging_as_socket_client()
            return pid
    except OSError, exc:
        exc_errno = exc.errno
        exc_strerror = exc.strerror
        error = ExecProcessError(
            "%(error_message)s: [%(exc_errno)d] %(exc_strerror)s" %
            vars())
        raise error


def create_example_logging_config(logging_config_filename):
    '''Write default daemon logging config.'''
    with open(logging_config_filename, 'w+') as logging_config:
        logging_config.write(textwrap.dedent('''
        [loggers]
        keys=root, picostack

        [handlers]
        keys=picostackHandler, consoleHandler

        [formatters]
        keys=trivial

        [logger_root]
        level=DEBUG
        handlers=

        [logger_picostack]
        level=DEBUG
        handlers=picostackHandler
        qualname=picostack
        propagate=0

        [handler_consoleHandler]
        class=StreamHandler
        level=DEBUG
        formatter=trivial
        args=(sys.stdout,)

        [handler_picostackHandler]
        class=handlers.RotatingFileHandler
        level=DEBUG
        formatter=trivial
        args=('%(logs_path)s/picostk_daemon.log', 'w')
        maxBytes: 5242880
        backupCount: 7

        ''' % {
            'logs_path': os.path.dirname(logging_config_filename),
        }) + textwrap.dedent('''
        [formatter_trivial]
        format=%(asctime)s - %(name)s - %(levelname)s - %(message)s
        datefmt=
        '''))



from django.db import models
from picostack.errors import DataModelError


VM_IN_CLONING = 'C'
# After cloning instance state is set to 'stopped'
VM_IS_STOPPED = 'S'
VM_IS_LAUNCHED = 'L'
VM_IS_RUNNING = 'R'
VM_HAS_FAILED = 'F'
VM_IS_TERMINATING = 'T'
VM_IS_TRASHED = 'W'
# If instance is removed, it has no state but is just deleted from the DB.

VM_STATES = (
    (VM_IN_CLONING, 'InCloning'),
    (VM_IS_STOPPED, 'Stopped'),
    (VM_IS_RUNNING, 'Running'),
    (VM_IS_LAUNCHED, 'Launched'),
    (VM_HAS_FAILED, 'Failed'),
    (VM_IS_TERMINATING, 'Terminating'),
    (VM_IS_TRASHED, 'Trashed'),
)

VM_PORTS = {
    'ssh': 22,
    'vnc': 5900,
    'rdp': 3389,
}

DEFAULT_FLAVOUR = 'tiny'


class VmImage(models.Model):

    name = models.CharField(max_length=60, unique=True)

    image_filename = models.CharField(max_length=120)

    # Used to check if we have enough free space when cloning (in MB).
    disk_size = models.PositiveIntegerField()

    def __repr__(self):
        return 'VM Image: <%s>' % self.name

    def __str__(self):
        return self.name


class Flavour(models.Model):

    name = models.CharField(max_length=60, unique=True)

    # in Megabytes
    memory_size = models.PositiveIntegerField(default=1024)

    # Number of cores
    num_of_cores = models.PositiveSmallIntegerField(default=1)

    def __repr__(self):
        return 'VM Flavour: <%s>' % self.name

    def __str__(self):
        return self.name


class VmInstance(models.Model):

    name = models.CharField(max_length=60, unique=True)

    image = models.ForeignKey(VmImage, related_name='instances')

    flavour = models.ForeignKey(Flavour, related_name='instances')

    current_state = models.CharField(
        max_length=1, choices=VM_STATES, default=VM_IN_CLONING)

    @property
    def memory_size(self):
        return self.flavour.memory_size

    @property
    def num_of_cores(self):
        return self.flavour.num_of_cores

    # has SSH ?
    has_ssh = models.BooleanField(default=False)

    # Set by vm_manager
    ssh_mapping = models.PositiveSmallIntegerField(null=True, blank=True)

    # has VNC ?
    has_vnc = models.BooleanField(default=False)

    # Set by vm_manager
    vnc_mapping = models.PositiveSmallIntegerField(null=True, blank=True)

    # has RDP ? Set True for Windows.
    has_rdp = models.BooleanField(default=False)

    # Set by vm_manager
    rdp_mapping = models.PositiveSmallIntegerField(null=True, blank=True)

    # Vnc port in '-vnc localhost:'
    localhost_vnc_port = models.PositiveSmallIntegerField(null=True,
                                                          blank=True)

    # If left blank, then the value from get_default_disk_filename() is used.
    disk_filename = models.CharField(max_length=120, null=True, blank=True)

    def change_state(self, state):
        self.current_state = state
        self.save(force_update=True)

    @staticmethod
    def get_all_occupied_ports():
        '''Get all ports occupied by running VM instances.'''
        port_mappings = list()
        for machine in VmInstance.objects.filter(current_state=VM_IS_RUNNING):
            ports = [
                machine.ssh_mapping,
                machine.vnc_mapping,
                machine.rdp_mapping,
            ]
            port_mappings.extend([port for port in ports
                                  if port is not None])
        return port_mappings

    def map_port(self, vm_port, host_port):
        if vm_port == 'ssh':
            assert self.has_ssh
            self.ssh_mapping = host_port
        elif vm_port == 'rdp':
            assert self.has_rdp
            self.rdp_mapping = host_port
        elif vm_port == 'vnc':
            assert self.has_vnc
            self.vnc_mapping = host_port
        else:
            raise Exception('Trying to map unknown port: %s' % vm_port)
        self.save(force_update=True)

    def get_default_disk_filename(self):
        return '%s_%s.dsk' % (self.image.image_filename, self.name)

    def get_default_localhost_vnc_port(self):
        allocated_ports = [machine.localhost_vnc_port for machine
                           in VmInstance.objects.all()]
        for port in range(1, VmInstance.objects.count()):
            if not port in allocated_ports:
                return port
        return VmInstance.objects.count() + 1

    def stop(self):
        # Reset/free all port mappings.
        self.ssh_mapping = None
        self.vnc_mapping = None
        self.rdp_mapping = None
        # Update state.
        self.current_state = 'Stopped'
        self.save(force_insert=True)

    @staticmethod
    def prepare_for_cloning(machine_name, vm_image, flavour):
        # Create a new instance.
        machine = VmInstance.objects.create(name=machine_name,
                                            image=vm_image,
                                            current_state='InCloning',
                                            flavour=flavour)
        machine.save()
        return machine

    @staticmethod
    def build_vm(vm_name, image_name, flavour_name=DEFAULT_FLAVOUR):
        vm_image = VmImage.objects.get(name=image_name)
        if not vm_image:
            raise DataModelError('VM image does not exists: %s' % image_name)
        flavour = Flavour.objects.get(name=flavour_name)
        if not flavour:
            raise DataModelError('VM flavour does not exists: %s' %
                                 flavour_name)
        if VmInstance.objects.filter(name=vm_name).exists():
            raise DataModelError('VM instance with this name already exists.')
        machine = VmInstance(
            name=vm_name,
            image=vm_image,
            flavour=flavour,
        )
        machine.save()

    # Some representation and casting implementation.
    def __repr__(self):
        return 'VM instance <%s> (flavour: %s, image: %s) ' % (
            self.name, self.flavour.name, self.image.name)

    def __str__(self):
        return '%s (%s, %s)' % (
            self.name, self.flavour.name, self.image.name)

    def save(self, *args, **kwargs):
        if self.disk_filename is None or self.disk_filename == '':
            self.disk_filename = self.get_default_disk_filename()
        if self.localhost_vnc_port is None or self.localhost_vnc_port == 0:
            self.localhost_vnc_port = self.get_default_localhost_vnc_port()
        super(VmInstance, self).save(*args, **kwargs)



#!/usr/bin/env python
# TODO: migrate to distribute?
#import distribute_setup
#distribute_setup.use_setuptools()


import os.path
import sys
from glob import glob

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def readme():
    try:
        with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
            return f.read()
    except (IOError, OSError):
        return ''


def get_version():
    src_path = os.path.join(os.path.dirname(__file__), 'picostack')
    sys.path.append(src_path)
    import picostack
    return picostack.__version__


setup(
    name='picostack',
    version=get_version(),
    description='A super lightweight KVM virtualization manager',
    long_description=readme(),
    author='Yauhen Yakimovich',
    author_email='eugeny.yakimovitch@gmail.com',
    url='https://github.com/ewiger/picostack',
    license='MIT',
    scripts=['picostk', 'picostk-django'],
    packages=[
        'picostack', 'picostack.vms', 'picostack.vms.templatetags',
    ],
    package_data={
        '': ['*.html', '*.svg', '*.js'],
    },
    include_package_data=True,
    download_url='https://github.com/ewiger/picostack/tarball/master',
    install_requires=[
        'sh >= 1.08',
        'daemoncxt >= 1.5.7',
        'Django >= 1.6.2',
        'psutil >= 2.1.1',
        'django-bootstrap3 >= 4.4.1',
    ],
    data_files=[
        ('/etc/init.d', ['pstk']),
    ],
    classifiers=[
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: System :: Emulators',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Framework :: Django',
        'Development Status :: 4 - Beta',
        'Operating System :: POSIX :: Linux',
    ],
)



from django.shortcuts import render
from django.http import HttpResponseRedirect, HttpResponse
from django import forms
from django.forms.models import modelformset_factory, ModelForm
from django.contrib.auth import logout
from django.contrib.auth.decorators import login_required
from picostack.vms.models import (VmInstance, VM_IS_LAUNCHED,
                                  VM_IS_TERMINATING, VM_IS_TRASHED)


class VmInstanceForm(ModelForm):
    class Meta:
        model = VmInstance
        fields = [
            'name', 'current_state', 'image', 'flavour',
            'has_ssh', 'has_rdp', 'has_vnc',
            #'ssh_mapping', 'rdp_mapping', 'vnc_mapping',
        ]
    #current_state = forms.ChoiceField(widget = forms.TextInput(
    #                                  attrs={'readonly':'readonly'}))


VmInstancesFormSet = modelformset_factory(model=VmInstance,
                                          form=VmInstanceForm,
                                          # No empty forms..
                                          extra=0)


def get_vm_instance(request, submit_id):
    # Initialize form set and pick the index.
    formset = VmInstancesFormSet(
        request.POST, request.FILES,
        queryset=VmInstance.objects.all(),
    )
    form_index = int(request.POST[submit_id][len(submit_id):]) - 1
    assert form_index < len(formset.forms) and form_index >= 0
    # Obtain corresponding model instance without saving anything.
    form = formset.forms[form_index]
    assert form.is_valid()
    return form.save(commit=False)


def get_view_context():
    # Instantiate form for each instance to pass to template.
    vm_instances_formset = VmInstancesFormSet()
    # Make list of column headers for the template.
    columns = list()
    if vm_instances_formset.total_form_count() > 0:
        columns = [field.label_tag for field
                   in vm_instances_formset.forms[0].visible_fields()]
    return {
        'formset': vm_instances_formset,
        'columns': columns,
    }

#
# Views
#
def logout_view(request):
    logout(request)
    return HttpResponseRedirect('/instances/')


def get_connection_details(request):
    if not request.GET.has_key('name'):
        raise Exception('Missing instance name')
    try:
        vm_instance = VmInstance.objects.get(name=request.GET['name'])
    except VmInstance.DoesNotExist:
        return HttpResponse('# Error. VM was not found by name: %s' %
                            request.GET['name'])
    hostname = request.get_host()
    if ':' in hostname:
        hostname = hostname[:hostname.index(':')]

    connection_str = 'ssh -T %s' % hostname
    port_mapping_template = '-L %d:localhost:%d'
    mappings = list()
    if vm_instance.has_vnc:
        mappings.append(port_mapping_template %
                        (vm_instance.vnc_mapping,
                         vm_instance.vnc_mapping))
    if vm_instance.has_ssh:
        mappings.append(port_mapping_template %
                        (vm_instance.ssh_mapping,
                         vm_instance.ssh_mapping))
    if vm_instance.has_rdp:
        mappings.append(port_mapping_template %
                        (vm_instance.rdp_mapping,
                         vm_instance.rdp_mapping))
    return HttpResponse(' '.join([connection_str] + mappings + ['\n']))


@login_required
def manage_instances(request):
    # TODO: handle errors
    # If the form has been submitted, handle the POST request per action.
    if request.method == 'POST':
        if '_save' in request.POST:
            formset = VmInstancesFormSet(
                request.POST, request.FILES,
                queryset=VmInstance.objects.all(),
            )
            if formset.is_valid():
                formset.save()  # FIXME: do we need to save?
        elif '_start' in request.POST:
            vm_instance = get_vm_instance(request, submit_id='_start')
            # Schedule VM for start.
            vm_instance.change_state(VM_IS_LAUNCHED)
        elif '_stop' in request.POST:
            vm_instance = get_vm_instance(request, submit_id='_stop')
            # Schedule VM for stop.
            vm_instance.change_state(VM_IS_TERMINATING)
        elif '_trash' in request.POST:
            vm_instance = get_vm_instance(request, submit_id='_trash')
            # Schedule VM for complete removal.
            vm_instance.change_state(VM_IS_TRASHED)
        return HttpResponseRedirect('/instances/')
    # Otherwise view instances. Render the template as response.
    return render(request, 'instances/view.html', get_view_context())


@login_required
def list_instances(request):
    # Render the template as response.
    context = get_view_context()
    context.update({
        'connect_url': request.build_absolute_uri('/connect_instance/'),
    })
    return render(request, 'instances/list.html', context)
